// Copyright (c) 2017, teja. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'package:jaguar_cache/jaguar_cache.dart';
import 'package:redis_client/redis_client.dart';
import 'package:jaguar_serializer/serializer.dart';

/// Cache implementation backed by redis database
class RedisCache extends Cache {
  /// Default expiry duration
  final Duration defaultExpiry;

  /// Connection string to redis db
  final String connectionString;

  /// Database number
  final int database;

  /// Repository to serialize/deserialize data
  final JsonRepo repo;

  /// Client
  RedisClient _client;

  RedisCache(this.defaultExpiry, this.repo,
      {this.connectionString, this.database});

  /// Connects to redis db
  Future _connect() async {
    if (_client == null) {
      _client = await RedisClient.connect(connectionString);
      _client.select(database);
    }
  }

  @override
  /// Set the given key/value in the cache, overwriting any existing value
  /// associated with that key
  Future upsert<T>(String key, T value, [Duration expires]) async {
    expires ??= defaultExpiry;
    await _connect();
    final String data = repo.serialize(
        {'e': new DateTime.now().add(expires).toIso8601String(), 'v': value},
        withType: true);
    await _client.set(key, data);
  }

  /// Set the given key/value in the cache ONLY IF the key already exists.
  @override
  Future replace<T>(String key, T v, [Duration expires]) async {
    await _connect();
    if (!await _client.exists(key)) return;
    await upsert(key, v, expires);
  }

  /// Deletes the given key from the cache
  @override
  Future remove<T>(String key) async {
    await _connect();
    await _client.del(key);
  }

  /// Get the content associated multiple keys at once
  @override
  Future<List<T>> readMany<T>(List<String> keys) async {
    await _connect();

    final ret = [];
    for (String key in keys) {
      ret.add(await read(key));
    }

    return ret;
  }

  /// Get the content associated with the given key
  @override
  Future<T> read<T>(String key) async {
    await _connect();
    final String v = await _client.get(key);

    if (v == null) throw cacheMiss;

    final Map map = repo.deserialize(v);
    final time = DateTime.parse(map['e']);

    if (new DateTime.now().isAfter(time)) throw cacheMiss;

    return map['v'];
  }
}
